#include "StdAfx.h"
#include "iplimagecapsule.h"
//#include <atlimage.h>

//#pragma comment(lib, "ipl")

//#include "cv.h"

IMPLEMENT_SERIAL(CIplImageCapsule, CObject, 1)

CIplImageCapsule::CIplImageCapsule(void)
	: m_bmi(NULL)
	, m_iplImage(NULL)
	, m_gsImage(NULL)
	, m_gsBmi(NULL)
	, m_parentDlg(NULL)
{
}

CIplImageCapsule::~CIplImageCapsule(void)
{
	Clear();
}

void CIplImageCapsule::Clear()
{
	if (m_bmi)
		free(m_bmi);
	if (m_gsBmi)
		free(m_gsBmi);

	if (m_iplImage)
	{
		if(m_iplImage->imageData)
			iplDeallocate(m_iplImage,IPL_IMAGE_ALL);
		else
			iplDeallocate(m_iplImage,IPL_IMAGE_HEADER);
	}
	if (m_gsImage)
	{
		if(m_gsImage->imageData)
			iplDeallocate(m_gsImage, IPL_IMAGE_ALL);
		else
			iplDeallocate(m_gsImage, IPL_IMAGE_HEADER);
	}

	m_bmi = NULL;
	m_gsBmi = NULL;
	m_iplImage = NULL;
	m_gsImage = NULL;
	m_parentDlg = NULL;
}

bool CIplImageCapsule::GetFromFile(CFile* file) 
{
	CIplImageCapsule::Clear();

	BITMAPFILEHEADER bmfh;
	file->Read(&bmfh,sizeof(bmfh));

	m_bmi = (BITMAPINFO*)malloc(bmfh.bfOffBits - sizeof(BITMAPFILEHEADER));
	file->Read(&m_bmi->bmiHeader, sizeof(BITMAPINFOHEADER));

	int colorsSize;

	colorsSize = bmfh.bfOffBits - sizeof(BITMAPFILEHEADER) - sizeof(BITMAPINFOHEADER);
	if (colorsSize > 0)
		file->Read(&m_bmi->bmiColors, colorsSize);

	int buffer_size;
	if (m_bmi->bmiHeader.biSizeImage==0)
		m_bmi->bmiHeader.biSizeImage = (((m_bmi->bmiHeader.biWidth*(m_bmi->bmiHeader.biBitCount)/8)+3)/4)*4*m_bmi->bmiHeader.biHeight;
	buffer_size = m_bmi->bmiHeader.biSizeImage;

	byte* buffer = new byte[buffer_size];

	file->Read(buffer, buffer_size);

	switch(m_bmi->bmiHeader.biCompression)
	{
	case BI_RGB:
		switch(m_bmi->bmiHeader.biBitCount)
		{
		case 24:
			m_iplImage = iplCreateImageHeader(3,0, IPL_DEPTH_8U,_T("RGB"),_T("BGR"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD,m_bmi->bmiHeader.biWidth,m_bmi->bmiHeader.biHeight, NULL,NULL,NULL,NULL);
			break;
		case 32:		
			m_iplImage = iplCreateImageHeader(3,1, IPL_DEPTH_8U,_T("RGB"),_T("BGRA"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD,m_bmi->bmiHeader.biWidth,m_bmi->bmiHeader.biHeight, NULL,NULL,NULL,NULL);
			break;
		case 8:			
			m_iplImage = iplCreateImageHeader(1,0, IPL_DEPTH_8U,_T("GRAY"),_T("GRAY"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD,m_bmi->bmiHeader.biWidth,m_bmi->bmiHeader.biHeight, NULL,NULL,NULL,NULL);
			break;
		default:		// not working properly
			delete buffer;
			return false;
			break;
		}
		break;
	default:
		delete buffer;
		return false;
		break;
	}
	iplConvertFromDIBSep((BITMAPINFOHEADER*)m_bmi, (const char*)buffer, m_iplImage);

	m_iplImage->origin = IPL_ORIGIN_TL;

	switch (m_iplImage->nChannels)
	{
	case 3: // RGB image has ben loaded -> convert it to grayscale
		m_gsImage = iplCreateImageHeader(1, 0, IPL_DEPTH_8U,_T("GRAY"),_T("GRAY"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD, m_iplImage->width, m_iplImage->height, NULL, NULL, NULL, NULL);
		iplAllocateImage(m_gsImage, FALSE, 0);
		iplColorToGray(m_iplImage, m_gsImage);

		m_gsBmi = (BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER)+ 256*sizeof(RGBQUAD));
		::ZeroMemory(&m_gsBmi->bmiHeader, sizeof(BITMAPINFOHEADER));
		m_gsBmi->bmiHeader.biSize        = sizeof(BITMAPINFOHEADER);
		m_gsBmi->bmiHeader.biWidth       = m_bmi->bmiHeader.biWidth;
		m_gsBmi->bmiHeader.biHeight      = m_bmi->bmiHeader.biHeight;
		m_gsBmi->bmiHeader.biCompression = BI_RGB;
		m_gsBmi->bmiHeader.biPlanes      = 1;
		m_gsBmi->bmiHeader.biBitCount	  = 8;
		for(int a = 0; a < 256; a++)
		{
			m_gsBmi->bmiColors[a].rgbRed = a;
			m_gsBmi->bmiColors[a].rgbGreen = a;
			m_gsBmi->bmiColors[a].rgbBlue = a;
		}
		break;
	case 1: // grayscale image has ben loaded -> make the default image RGB
		m_gsImage = m_iplImage;
		m_iplImage = iplCreateImageHeader(3, 0, IPL_DEPTH_8U,_T("RGB"),_T("BGR"), IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD, m_gsImage->width, m_gsImage->height, NULL, NULL, NULL, NULL);
		iplAllocateImage(m_iplImage, FALSE, 0);
		iplGrayToColor(m_gsImage, m_iplImage, 1, 1, 1);

		m_gsBmi = m_bmi;
		m_bmi = (BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER));
		::ZeroMemory(&m_bmi->bmiHeader, sizeof(BITMAPINFOHEADER));
		m_bmi->bmiHeader.biSize        = sizeof(BITMAPINFOHEADER);
		m_bmi->bmiHeader.biWidth       = m_gsBmi->bmiHeader.biWidth;
		m_bmi->bmiHeader.biHeight      = m_gsBmi->bmiHeader.biHeight;
		m_bmi->bmiHeader.biCompression = BI_RGB;
		m_bmi->bmiHeader.biPlanes      = 1;
		m_bmi->bmiHeader.biBitCount	  = 24;
		break;
	default:
		delete buffer;
		return false;
	}

	delete buffer;
	return true;
}

bool CIplImageCapsule::GetFromFile(CString fileName)
{
	CFile f;
	if (!f.Open(fileName, CFile::shareDenyWrite))
		return false;
	bool result = GetFromFile(&f);
	f.Close();
	return result;
}


void CIplImageCapsule::DrawToDC(CZoomDC* dc, bool color)
{
	if (color)
		SetDIBitsToDevice(*(dc->m_dc)
			, -dc->m_shift.cx + dc->m_topLeftOffset.cx, -dc->m_shift.cy  + dc->m_topLeftOffset.cy 
			, m_iplImage->width, m_iplImage->height
			, 0, 0
			, 0, m_iplImage->height
			
			, m_iplImage->imageData, m_bmi, 
		DIB_RGB_COLORS);
	else
		SetDIBitsToDevice(*(dc->m_dc)
			, -dc->m_shift.cx + dc->m_topLeftOffset.cx, -dc->m_shift.cy  + dc->m_topLeftOffset.cy
			, m_gsImage->width, m_gsImage->height
			, 0, 0
			, 0, m_gsImage->height
			
			, m_gsImage->imageData, m_gsBmi, 
		DIB_RGB_COLORS);
}

void CIplImageCapsule::StretchToDC(CZoomDC* dc, bool color)
{
	CRect clipBox;
	CRect imageBox;
	CRect imageRect;
	imageRect.SetRect(0,0,m_bmi->bmiHeader.biWidth, m_bmi->bmiHeader.biHeight); 
	dc->ImageToView(imageRect);
	dc->m_dc->GetClipBox(&clipBox);
	clipBox.IntersectRect(clipBox, imageRect);
	imageBox = clipBox;
	dc->ViewToImage(imageBox);
	imageBox.InflateRect(max(1, dc->m_zoom), max(1, dc->m_zoom));
	imageBox.left = max(0, imageBox.left);
	imageBox.top = max(0, imageBox.top);
	imageBox.right = min(m_bmi->bmiHeader.biWidth - 1, imageBox.right);
	imageBox.bottom = min(m_bmi->bmiHeader.biHeight - 1, imageBox.bottom);
	clipBox = imageBox;
	dc->ImageToView(clipBox);
	if(dc->m_zoom<0)
		dc->m_dc->SetStretchBltMode(STRETCH_HALFTONE);
	else
		dc->m_dc->SetStretchBltMode(STRETCH_DELETESCANS);
	if (color)
	{
		if (m_iplImage->origin == IPL_ORIGIN_TL)
			::StretchDIBits(*(dc->m_dc)
			, clipBox.left, clipBox.bottom - 1
			,  clipBox.Width(), -clipBox.Height()
			, imageBox.left, imageBox.top
			, imageBox.Width(), imageBox.Height()
				
				, m_iplImage->imageData, m_bmi, DIB_RGB_COLORS, SRCCOPY);
		else
			::StretchDIBits(*(dc->m_dc)
			, clipBox.left, clipBox.top
			,  clipBox.Width(), clipBox.Height()
			, imageBox.left, imageBox.top
			, imageBox.Width(), imageBox.Height()
				
				, m_iplImage->imageData, m_bmi, DIB_RGB_COLORS, SRCCOPY);
	}
	else
	{
		if (m_gsImage->origin == IPL_ORIGIN_TL)
			::StretchDIBits(*(dc->m_dc)
				, clipBox.left, clipBox.bottom - 1
				,  clipBox.Width(), -clipBox.Height()
				, imageBox.left, imageBox.top
				, imageBox.Width(), imageBox.Height()
				
				, m_gsImage->imageData, m_gsBmi, DIB_RGB_COLORS, SRCCOPY);
		else
			::StretchDIBits(*(dc->m_dc)
				, clipBox.left, clipBox.top
				,  clipBox.Width(), clipBox.Height()
				, imageBox.left, imageBox.top
				, imageBox.Width(), imageBox.Height()
				
				, m_gsImage->imageData, m_gsBmi, DIB_RGB_COLORS, SRCCOPY);
	}
}

void CIplImageCapsule::InitBitmapInfos(int width, int height, bool initGSImage)
{
	if (m_bmi)
		free(m_bmi);
	if (m_gsBmi)
		free(m_gsBmi);

	m_bmi = NULL;
	m_gsBmi = NULL;

	if (initGSImage)
	{
		m_gsBmi = (BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER)+ 256*sizeof(RGBQUAD));
		::ZeroMemory(&m_gsBmi->bmiHeader, sizeof(BITMAPINFOHEADER));
		m_gsBmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
		m_gsBmi->bmiHeader.biWidth = width;
		m_gsBmi->bmiHeader.biHeight = height;
		m_gsBmi->bmiHeader.biCompression = BI_RGB;
		m_gsBmi->bmiHeader.biPlanes = 1;
		m_gsBmi->bmiHeader.biBitCount = 8;
		for (int a = 0; a < 256; a++)
		{
			m_gsBmi->bmiColors[a].rgbRed = a;
			m_gsBmi->bmiColors[a].rgbGreen = a;
			m_gsBmi->bmiColors[a].rgbBlue = a;
		}
	}

	m_bmi = (BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER));
	::ZeroMemory(&m_bmi->bmiHeader, sizeof(BITMAPINFOHEADER));
	m_bmi->bmiHeader.biBitCount = 24;
	m_bmi->bmiHeader.biCompression = BI_RGB;
	m_bmi->bmiHeader.biPlanes = 1;
	m_bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	m_bmi->bmiHeader.biWidth = width;
	m_bmi->bmiHeader.biHeight = height;
}

void CIplImageCapsule::InitImages(int width, int height, bool allocateImage, bool initGSImage)
{
	if (m_iplImage)
	{
		if(!m_allocatedImageData)
			m_iplImage->imageData = NULL;
		iplDeallocate(m_iplImage, IPL_IMAGE_ALL);
	}
	m_iplImage = iplCreateImageHeader(3, 0, IPL_DEPTH_8U,_T("RGB"),_T("BGR"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD, width, height, NULL,NULL,NULL,NULL);
	if (allocateImage)
		iplAllocateImage(m_iplImage, TRUE, 0);

	if (m_gsImage)
	{
		if(!m_allocatedImageData)
			m_gsImage->imageData = NULL;
		iplDeallocate(m_gsImage, IPL_IMAGE_ALL);
		m_gsImage = NULL;
	}

	if (initGSImage)
	{
		m_gsImage = iplCreateImageHeader(1,0, IPL_DEPTH_8U,_T("GRAY"),_T("GRAY"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD, width, height, NULL, NULL, NULL, NULL);
		if (allocateImage)
			iplAllocateImage(m_gsImage, TRUE, 0);
	}

	m_allocatedImageData = allocateImage;

}

void CIplImageCapsule::Serialize(CArchive& archive)
{
	CObject::Serialize(archive);

	if (archive.IsStoring())
	{
		if (m_iplImage)
		{
			archive << m_iplImage->imageSize << m_iplImage->width << m_iplImage->height;
			archive.Write((BYTE*)m_iplImage->imageData, m_iplImage->imageSize);
		}
		else
			archive << 0;

		if (m_gsImage)
		{
			archive << m_gsImage->imageSize << m_gsImage->width << m_gsImage->height;
			archive.Write((BYTE*)m_gsImage->imageData, m_gsImage->imageSize);
		}
		else
			archive << 0;
	}
	else
	{
		Clear();

		int imageSize, width, height;
		archive >> imageSize;
		if (imageSize > 0)
		{
			archive >> width >> height;

			ASSERT(width);
			ASSERT(height);
			InitBitmapInfos(width, height);
			InitImages(width, height);
			archive.Read((BYTE*)m_iplImage->imageData, imageSize);
		}

		archive >> imageSize;
		if (imageSize > 0)
		{
			archive >> width >> height;

			ASSERT(width);
			ASSERT(height);
			archive.Read((BYTE*)m_gsImage->imageData, imageSize);
		}
	}
}

int CIplImageCapsule::GetFromIplImage(IplImage* iplImage)
{
	CIplImageCapsule::Clear();
	if(iplImage)
	{
		InitBitmapInfos(iplImage->width,iplImage->height);
		switch(iplImage->nChannels)
		{
		case 3:
			m_iplImage = iplCloneImage(iplImage);
			m_gsImage = iplCreateImageHeader(1, 0, IPL_DEPTH_8U,_T("GRAY"),_T("GRAY"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD, iplImage->width, iplImage->height, NULL, NULL, NULL, NULL);
			iplAllocateImage(m_gsImage, FALSE, 0);
			iplColorToGray(m_iplImage, m_gsImage);
			break;
		case 4:
			m_iplImage = iplCreateImageHeader(3, 0, IPL_DEPTH_8U,_T("RGB"),_T("BGR"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD,iplImage->width,iplImage->height, NULL,NULL,NULL,NULL);
			iplAllocateImage(m_iplImage, FALSE, 0);
//			iplConvert(iplImage, m_iplImage);
			{
				IplROI* roi = iplCreateROI(0,0,0,iplImage->width,iplImage->height);
				m_iplImage->roi = roi;
				iplImage->roi = roi;
				for(int channelID = 1; channelID < 4; channelID++)
				{
					roi->coi = channelID;
					iplCopy(iplImage, m_iplImage);
				}
				m_iplImage->roi = NULL;
				iplImage->roi = NULL;
				iplDeleteROI(roi);
			}
			m_gsImage = iplCreateImageHeader(1, 0, IPL_DEPTH_8U,_T("GRAY"),_T("GRAY"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD, iplImage->width, iplImage->height, NULL, NULL, NULL, NULL);
			iplAllocateImage(m_gsImage, FALSE, 0);
			iplColorToGray(m_iplImage, m_gsImage);
			break;
		case 1:			
			m_gsImage = iplCloneImage(iplImage);
			m_iplImage = iplCreateImageHeader(3, 0, IPL_DEPTH_8U,_T("RGB"),_T("BGR"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD,iplImage->width,iplImage->height, NULL,NULL,NULL,NULL);
			iplAllocateImage(m_iplImage, FALSE, 0);
			iplGrayToColor(iplImage, m_iplImage, 1, 1, 1);
			break;
		default:		// not working properly
			ASSERT(0);
			return 0;
			break;
		}
	}
	return 1;
}

bool CIplImageCapsule::SaveToFile(CString filename, bool colorImage)
{
	BITMAPFILEHEADER hdr;

	FILE *fw;
	IplImage* img;

	fw = fopen(filename, "wb");
	if (fw == NULL)
		return false;
	hdr.bfType = 'MB';
	if(colorImage)
	{
		img = iplCloneImage(m_iplImage);
		img->origin = IPL_ORIGIN_BL;
		iplConvert(m_iplImage,img);
		hdr.bfSize = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+img->imageSize;
	}
	else
	{
		img = iplCloneImage(m_gsImage);
		img->origin = IPL_ORIGIN_BL;
		iplConvert(m_gsImage,img);
		hdr.bfSize = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+ 256 * sizeof(RGBQUAD) + img->imageSize;
	}
	hdr.bfReserved1 = hdr.bfReserved2 = 0;
	hdr.bfOffBits = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);
	if(!colorImage)
		hdr.bfOffBits += 256 * sizeof(RGBQUAD);
	fwrite(&hdr, sizeof(BITMAPFILEHEADER), 1, fw);
	if(colorImage)
	{
		fwrite(m_bmi, 1, sizeof(BITMAPINFOHEADER), fw);
		fwrite(img->imageData, 1, img->imageSize, fw);
	}
	else
	{
		fwrite(m_gsBmi, 1, sizeof(BITMAPINFOHEADER) + 256*sizeof(RGBQUAD), fw);
		fwrite(img->imageData, 1, img->imageSize, fw);
	}
	fclose(fw);
	iplDeallocate(img,IPL_IMAGE_ALL);
	return true;
}
