// IplImageLoader.cpp : implementation file
//

#include "stdafx.h"
#include "IplImageLoader.h"

// CIplImageLoader

CIplImageLoader::CIplImageLoader()
	: m_iplImage(NULL)
	, m_pBitmap(NULL)
	, m_isImageLoaded(false)
	, m_bmi(NULL)
	, m_dibData(NULL)
{
	if (m_bmi)
		free(m_bmi);

	m_bmi = (BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER));
	::ZeroMemory(&m_bmi->bmiHeader, sizeof(BITMAPINFOHEADER));
	m_bmi->bmiHeader.biBitCount = 24;
	m_bmi->bmiHeader.biCompression = BI_RGB;
	m_bmi->bmiHeader.biPlanes = 1;
	m_bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
// 	m_bmi->bmiHeader.biWidth = width;
// 	m_bmi->bmiHeader.biHeight = height;
}

CIplImageLoader::~CIplImageLoader()
{
	if (m_isImageLoaded)
	{
		if (m_iplImage)
		{
			iplDeallocate(m_iplImage, IPL_IMAGE_ALL);
			m_iplImage = NULL;
		}
	}

	if (m_pBitmap)
	{
		delete m_pBitmap;
		m_pBitmap = NULL;
	}

	if (m_bmi)
	{
		delete m_bmi;
		m_bmi = NULL;
	}

	if (m_dibData)
	{
		delete m_dibData;
		m_dibData = NULL;
	}
}
// CIplImageLoader member functions

bool CIplImageLoader::IsImageLoaded()
{
	return m_isImageLoaded;
}

bool CIplImageLoader::LoadImage(CString filename)
{
	CFile file;
	bool retVal = false;

	if (filename.IsEmpty())
		return false;

	if (file.Open(filename, CFile::modeRead))
	{
		int imageSize = (int)file.GetLength();
		void* imageData = new byte[imageSize];

		if (imageData)
		{
			file.Read(imageData, imageSize);
		}
		else
		{
			file.Close();
			return false;
		}

		file.Close();

		HGLOBAL hBuffer = ::GlobalAlloc(GMEM_MOVEABLE, imageSize);

		if (hBuffer)
		{
			void* pBuffer = ::GlobalLock(hBuffer);

			if (pBuffer)
			{
				CopyMemory(pBuffer, imageData, imageSize);

				IStream* pStream = NULL;

				if (::CreateStreamOnHGlobal(hBuffer, FALSE, &pStream) == S_OK)
				{
					if (m_pBitmap)
					{
						delete m_pBitmap;
						m_pBitmap = NULL;
					}

					m_pBitmap = Gdiplus::Bitmap::FromStream(pStream);
					pStream->Release();

					if (m_pBitmap)
					{ 
						if (m_pBitmap->GetLastStatus() == Gdiplus::Ok)
						{
							if (imageData)
							{
								delete imageData;
								imageData = NULL;
							}

							retVal = true;
						}
						else
						{
							delete m_pBitmap;
							m_pBitmap = NULL;
						}
					}
				}

				if (!retVal)
				{
					m_pBitmap = NULL;
					::GlobalUnlock(hBuffer);
				}
			}
			if (!retVal)
			{
				::GlobalFree(hBuffer);
				hBuffer = NULL;
			}
		}

		if (imageData)
		{
			delete imageData;
			imageData = NULL;
		}
	}

	m_isImageLoaded = retVal;

	if (m_isImageLoaded)
	{
		BmpToIpl();
	}

	return retVal;
}

void CIplImageLoader::IplToBmp()
{
	if (!m_iplImage)
		return;

	if (m_pBitmap)
	{
		delete m_pBitmap;
		m_pBitmap = NULL;
	}

	BITMAPINFO* dibHeader;
	/*	::ZeroMemory(&dibHeader,sizeof(BITMAPINFO));*/
	if (m_iplImage->nChannels == 3)
	{
		dibHeader = (BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER));
	}
	else
	{
		dibHeader = (BITMAPINFO*)malloc(sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD));
		for(int i =0; i<256;i++)
		{
			dibHeader->bmiColors[i].rgbRed = i;
			dibHeader->bmiColors[i].rgbGreen = i;
			dibHeader->bmiColors[i].rgbBlue = i;
			dibHeader->bmiColors[i].rgbReserved = 0;
		}
	}

	BITMAPINFOHEADER* infoHeader = &dibHeader->bmiHeader;

	::ZeroMemory(infoHeader, sizeof(BITMAPINFOHEADER));
	infoHeader->biHeight = m_iplImage->height;
	infoHeader->biWidth = m_iplImage->width;	
	infoHeader->biBitCount = m_iplImage->nChannels*8;
	infoHeader->biSizeImage = (((infoHeader->biWidth*(infoHeader->biBitCount)/8)+3)/4)*4*infoHeader->biHeight;
	infoHeader->biSize = sizeof(BITMAPINFOHEADER);
	infoHeader->biCompression = BI_RGB;
	infoHeader->biPlanes = 1;

	if (m_dibData)
	{
		delete m_dibData;
		m_dibData = NULL;
	}

	m_dibData = new char[m_iplImage->imageSize];
	iplConvertToDIBSep(m_iplImage, (BITMAPINFOHEADER*)dibHeader, m_dibData, IPL_DITHER_NONE, IPL_PALCONV_NONE);
	m_pBitmap = Bitmap::FromBITMAPINFO(dibHeader, (VOID*)m_dibData);

	free(dibHeader);
	dibHeader = NULL;
	infoHeader = NULL;
}

void CIplImageLoader::BmpToIpl()
{
	if (m_iplImage)
	{
		iplDeallocate(m_iplImage, IPL_IMAGE_ALL);
		m_iplImage = NULL;
	}

	if (!m_pBitmap)
		return;

	BITMAPINFOHEADER infoHeader;
	::ZeroMemory(&infoHeader, sizeof(BITMAPINFOHEADER));

	PixelFormat pixelFormat = m_pBitmap->GetPixelFormat();
	switch(pixelFormat) {
		case PixelFormat8bppIndexed:
			m_iplImage = iplCreateImageHeader(1, 0,IPL_DEPTH_8U, "GRAY", "GRAY", IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL,IPL_ALIGN_4BYTES, m_pBitmap->GetWidth(), m_pBitmap->GetHeight(), NULL, NULL, NULL, NULL);
			break;
		case PixelFormat24bppRGB :
			m_iplImage = iplCreateImageHeader(3, 0, IPL_DEPTH_8U, _T("RGB"), _T("BGR"), IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD, m_pBitmap->GetWidth(), m_pBitmap->GetHeight(), NULL, NULL, NULL, NULL);
			break;
		case PixelFormat32bppRGB:
			m_iplImage = iplCreateImageHeader(3, 0, IPL_DEPTH_8U, _T("RGB"), _T("BGR"), IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD, m_pBitmap->GetWidth(), m_pBitmap->GetHeight(), NULL, NULL, NULL, NULL);
			break;
		case PixelFormat32bppARGB:
			m_iplImage = iplCreateImageHeader(4, 4, IPL_DEPTH_8U,_T("RGBA"),_T("BGRA"),IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, IPL_ALIGN_DWORD,m_pBitmap->GetWidth(),m_pBitmap->GetHeight(), NULL,NULL,NULL,NULL);
			break;
		case PixelFormat1bppIndexed:
		case PixelFormat4bppIndexed:
		case PixelFormat16bppGrayScale:
		case PixelFormat16bppRGB555:
		case PixelFormat16bppRGB565:
		case PixelFormat16bppARGB1555:
		case PixelFormat32bppPARGB:
		case PixelFormat48bppRGB:
		case PixelFormat64bppARGB:
		case PixelFormat64bppPARGB:
			m_iplImage = NULL;
			AfxMessageBox("Format not supported!", MB_ICONERROR);
		default:
			ASSERT(FALSE);
			break;
	}

	infoHeader.biHeight = m_pBitmap->GetHeight();
	infoHeader.biWidth = m_pBitmap->GetWidth();	
	infoHeader.biSizeImage = (((infoHeader.biWidth*(infoHeader.biBitCount)/8)+3)/4)*4*infoHeader.biHeight;
	infoHeader.biSize = sizeof(BITMAPINFOHEADER);
	infoHeader.biCompression = BI_RGB;
	infoHeader.biPlanes = 1;

	HBITMAP hbmp;
	if (m_iplImage && m_pBitmap->GetHBITMAP(Color(255, 255, 255), &hbmp) == 0)
	{

		BITMAP bbb;
		GetObject((HGDIOBJ)hbmp, sizeof(BITMAP), &bbb);
		infoHeader.biBitCount = bbb.bmBitsPixel;
		iplConvertFromDIBSep(&infoHeader, (char*)bbb.bmBits, m_iplImage);
	}

	m_bmi->bmiHeader.biWidth = m_iplImage->width;
	m_bmi->bmiHeader.biHeight = -m_iplImage->height;
}

bool CIplImageLoader::DrawImage(CDC* dc, IplImage* image, int x, int y)
{
	if (image)
	{
		UpdateBitmapInfo(image);

		int retValue = SetDIBitsToDevice(dc->GetSafeHdc(),
			x, y, image->width, image->height,
			0, 0, 0, image->height,
			image->imageData, m_bmi, DIB_RGB_COLORS);

		return (retValue != 0);
	}
	else if (m_isImageLoaded)
	{
		UpdateBitmapInfo(m_iplImage);

		int retValue = SetDIBitsToDevice(dc->GetSafeHdc(),
			x, y, m_iplImage->width, m_iplImage->height,
			0, 0, 0, m_iplImage->height,
			m_iplImage->imageData, m_bmi, DIB_RGB_COLORS);

		return (retValue != 0);
	}

	return false;
}

bool CIplImageLoader::StretchImage(CDC* dc, CRect rect, IplImage* image)
{
	if (image)
	{
		UpdateBitmapInfo(image);

		int retValue = ::StretchDIBits(dc->GetSafeHdc(),
			rect.left, rect.top, rect.Width(), rect.Height(),
			0, 0, image->width, image->height,
			image->imageData, m_bmi, DIB_RGB_COLORS, SRCCOPY);

		return (retValue != 0);
	}
	else if (m_isImageLoaded)
	{
		//dc->SetStretchBltMode(STRETCH_DELETESCANS);
		//dc->SetStretchBltMode(STRETCH_HALFTONE);

		UpdateBitmapInfo(m_iplImage);

		int retValue = ::StretchDIBits(dc->GetSafeHdc(),
			rect.left, rect.top, rect.Width(), rect.Height(),
			0, 0, m_iplImage->width, m_iplImage->height,
			m_iplImage->imageData, m_bmi, DIB_RGB_COLORS, SRCCOPY);

		return (retValue != 0);
	}

	return false;
}

IplImage* CIplImageLoader::GetIplImage()
{
	return m_iplImage;
}

void CIplImageLoader::ResetIplImage(IplImage* iplImage)
{
	if (m_isImageLoaded)
	{
		if (m_iplImage)
		{
			iplDeallocate(m_iplImage, IPL_IMAGE_ALL);
			m_iplImage = NULL;
		}

		m_isImageLoaded = false;
	}

	m_iplImage = iplImage;

	//UpdateBitmapInfo(m_iplImage);
}

bool CIplImageLoader::SaveImage(CString filename, CString format)
{
	CFile outputFile;
	bool retVal = false;
	ULONG quality = 50;
	CString formatString = GetFormatString(format);

	if (m_iplImage)
		IplToBmp();

	if (outputFile.Open(filename, CFile::modeWrite | CFile::modeCreate))
	{
		if (!m_pBitmap)
			return false;

		Bitmap* image = m_pBitmap;

		// Create stream with 0 size
		IStream* pIStream    = NULL;
		if (CreateStreamOnHGlobal(NULL, TRUE, (LPSTREAM*)&pIStream) != S_OK)
		{
			AfxMessageBox(_T("Failed to create stream on global memory!"));
			return false;
		}

		CLSID pngClsid;
		WCHAR formatStringWide[256];
		MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, formatString.GetBuffer(0), formatString.GetLength(), formatStringWide, 255);
		formatStringWide[formatString.GetLength()]='\0';

		GetEncoderClsid(formatStringWide, &pngClsid);

		// Setup encoder parameters
		EncoderParameters encoderParameters;
		encoderParameters.Count = 1;
		encoderParameters.Parameter[0].Guid = EncoderQuality;
		encoderParameters.Parameter[0].Type = EncoderParameterValueTypeLong;
		encoderParameters.Parameter[0].NumberOfValues = 1;

		// setup compression level
		encoderParameters.Parameter[0].Value = &quality;

		//  Save the image to the stream
		Status SaveStatus = image->Save(pIStream, &pngClsid, &encoderParameters);
		if (SaveStatus != Ok)
		{
			// this should free global memory used by the stream
			// according to MSDN
			pIStream->Release();
			AfxMessageBox(_T("Failed to save to stream!"));
			return false;
		}

		// get the size of the stream
		ULARGE_INTEGER ulnSize;
		LARGE_INTEGER lnOffset;
		lnOffset.QuadPart = 0;
		if(pIStream->Seek(lnOffset, STREAM_SEEK_END, &ulnSize) != S_OK)
		{
			pIStream->Release();
			AfxMessageBox(_T("Failed to get the size of the stream!"));
			return false;
		}

		// now move the pointer to the beginning of the file
		if(pIStream->Seek(lnOffset, STREAM_SEEK_SET, NULL) != S_OK)
		{
			pIStream->Release();
			AfxMessageBox(_T("Failed to move the file pointer to the beginning of the stream!"));
			return false;
		}

		char *pBuff = new char[(int)ulnSize.QuadPart];
		// Read the stream directly into the buffer
		ULONG ulBytesRead;
		if (pIStream->Read(pBuff, (int)ulnSize.QuadPart, &ulBytesRead) != S_OK)
		{
			pIStream->Release();
			delete pBuff;
			return false;
		}

		outputFile.Write(pBuff, ulBytesRead);

		delete pBuff;

		// Free memory used by the stream
		pIStream->Release();

		outputFile.Close();
	}

	return true;
}

CString CIplImageLoader::GetFormatString(CString fileFormat)
{
	CString formatString;

	if (fileFormat.CompareNoCase("bmp")==0)
	{
		formatString = "image/bmp";
	}

	if (fileFormat.CompareNoCase("jpg")==0)
	{
		formatString = "image/jpeg";
	}

	if (fileFormat.CompareNoCase("gif")==0)
	{
		formatString = "image/gif";
	}

	if (fileFormat.CompareNoCase("tiff")==0)
	{
		formatString = "image/tiff";
	}

	if (fileFormat.CompareNoCase("png")==0)
	{
		formatString = "image/png";
	}

	return formatString;
}

int CIplImageLoader::GetEncoderClsid(const WCHAR* format, CLSID* pClsid)
{
	UINT  num = 0;          // number of image encoders
	UINT  size = 0;         // size of the image encoder array in bytes

	ImageCodecInfo* pImageCodecInfo = NULL;

	GetImageEncodersSize(&num, &size);
	if (size == 0)
		return -1;  // Failure

	pImageCodecInfo = new ImageCodecInfo[size];
	if (pImageCodecInfo == NULL)
		return -1;  // Failure

	GetImageEncoders(num, size, pImageCodecInfo);

	for (UINT j = 0; j < num; ++j)
	{
		if (wcscmp(pImageCodecInfo[j].MimeType, format) == 0)
		{
			*pClsid = pImageCodecInfo[j].Clsid;
			delete pImageCodecInfo;
			pImageCodecInfo = NULL;

			return j;  // Success
		}   
	}

	if (pImageCodecInfo)
	{
		delete pImageCodecInfo;
		pImageCodecInfo = NULL;
	}

	return -1;  // Failure
}

void CIplImageLoader::UpdateBitmapInfo(IplImage* image)
{
	if (image)
	{
		::ZeroMemory(&m_bmi->bmiHeader, sizeof(BITMAPINFOHEADER));
		m_bmi->bmiHeader.biBitCount = image->depth * image->nChannels;
		m_bmi->bmiHeader.biCompression = BI_RGB;
		m_bmi->bmiHeader.biPlanes = 1;
		m_bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
		m_bmi->bmiHeader.biWidth = image->width;
		m_bmi->bmiHeader.biHeight = -image->height;
	}
}